from .client import APIClient, ProductionAPIClient, DevelopmentAPIClient  # noqa: F401
from .exceptions import OnboardApiException, OnboardTemporaryException  # noqa: F401

OnboardClient = ProductionAPIClient



import pandas as pd


def points_df_from_timeseries(timeseries, points=[]) -> pd.DataFrame:
    """Returns a pandas dataframe from the results of a timeseries query"""

    # 'type' is from the point_type.display_name column
    point_names = {str(p['id']): p.get('type') for p in points}

    columns = ['timestamp']
    dates = set()
    data_by_point = {}

    for point in timeseries:
        point_id = point['tags']['point_id']
        columns.append(point_id)
        col_indexes = point['columns']

        ts_index = col_indexes.index('time')
        clean_index = col_indexes.index('clean')

        point_data = {}
        data_by_point[point_id] = point_data

        for val in point['values']:
            ts = val[ts_index]
            dates.add(ts)
            clean = val[clean_index]
            point_data[ts] = clean

    dates = list(dates)
    dates.sort()
    data = []

    for d in dates:
        row = {'timestamp': d}
        for p in columns[1:]:
            val = data_by_point[p].get(d)
            point_name = point_names.get(p)
            point_col = f"{point_name} - {p}" if point_name else p
            row[point_col] = val
        data.append(row)

    df = pd.DataFrame(data)
    return df



from .exceptions import OnboardApiException, OnboardTemporaryException


def divide_chunks(l, n):
    # looping till length l
    for i in range(0, len(l), n):
        yield l[i:i + n]


def json(func):
    """Decorator for making sure requests responses are handled consistently"""
    def wrapper(*args, **kwargs):
        try:
            res = func(*args, **kwargs)
            if res is None:
                return None

            # remove the cached access token if authorization failed
            # it's likely just expired
            if res.status_code == 401 and args[0].token is not None:
                args[0].token = None
                return wrapper(*args, **kwargs)

            if res.status_code > 499:
                raise OnboardTemporaryException(res.text)
            if res.status_code > 399:
                raise OnboardApiException(res.text)

            if hasattr(wrapper, 'raw_response'):
                return res
            return res.json()
        except OnboardApiException as e:
            raise e
        except Exception as e:
            raise OnboardApiException(e)
    return wrapper



import requests
import datetime
from requests.models import Response
from typing import Optional, Union
from .exceptions import OnboardApiException
from .util import json

USER_AGENT = 'Onboard Py-SDK'


class ClientBase:
    """Base class that implements HTTP methods against the API on top of requests"""
    def __init__(self, api_url: Optional[str],
                 user: Optional[str], pw: Optional[str],
                 api_key: Optional[str],
                 token: Optional[str],
                 name: [Optional[str]]) -> None:
        self.api_url = api_url
        self.api_key = api_key
        self.user = user
        self.pw = pw
        self.token = token
        self.name = name
        if not (api_key or token or (user and pw)):
            raise OnboardApiException("Need one of: user & pw, token or api_key")
        self.session = None

    def __session(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update(self.headers())
            self.session.headers.update(self.auth())
        return self.session

    def headers(self):
        agent = f"{USER_AGENT} ({self.name})" if self.name else USER_AGENT
        return {'Content-Type': 'application/json',
                'User-Agent': agent}

    def auth(self):
        token = self.__get_token()
        return {'Authorization': f'Bearer {token}'}

    @json
    def __pw_login(self):
        payload = {
            'login': self.user,
            'password': self.pw,
        }
        return self.post('/login', json=payload)

    @json
    def __api_key_login(self):
        payload = {'key': self.api_key}
        return self.post('/login/api-key', json=payload)

    def __get_token(self):
        if self.token is None:
            if self.api_key:
                login_res = self.__api_key_login()
            else:
                login_res = self.__pw_login()
            self.token = login_res['access_token']

        if self.token is None:
            raise OnboardApiException("Not authorized")

        return self.token

    def __repr__(self) -> str:
        return f"OnboardSdk(url={self.api_url})"

    def url(self, url: str) -> str:
        if not url.startswith('http'):
            return self.api_url + url
        return url

    def get(self, url: str, **kwargs) -> Response:
        return self.__session().get(self.url(url), **kwargs)

    def delete(self, url: str, **kwargs) -> Response:
        return self.__session().delete(self.url(url), **kwargs)

    def put(self, url: str, **kwargs) -> Response:
        return self.__session().put(self.url(url), **kwargs)

    def post(self, url: str, **kwargs) -> Response:
        return self.__session().post(self.url(url), **kwargs)

    def ts_to_dt(self, ts: Optional[float]) -> datetime.datetime:
        if ts is None:
            return None
        return datetime.datetime.utcfromtimestamp(ts / 1000.0)

    def dt_to_str(self, dt: Union[str, datetime.datetime]) -> str:
        if isinstance(dt, str):
            return dt
        if dt.tzinfo is None:
            return dt.isoformat() + "Z"
        return dt.isoformat()



class OnboardApiException(Exception):
    """Wrapper for exceptions throw by the API client"""
    pass


class OnboardTemporaryException(OnboardApiException):
    """These exceptions indicate that a call failed in a temporary manner
    and should be retried
    """
    pass



import math
from datetime import datetime, timezone
from typing import List, Optional, Union
from dataclasses import field
from pydantic.dataclasses import dataclass
from pydantic import validator


class PointDataUpdate(object):
    """Model for bulk-updating a point's data value and timestamp"""
    __slots__ = ['point_id', 'value', 'last_updated']

    def __init__(self, point_id, value, last_updated):
        errors = []
        if not isinstance(point_id, int):
            errors.append(f"point id must be an integer, saw {point_id}")
        self.point_id = point_id
        self.value = value
        if not isinstance(last_updated, datetime):
            errors.append(f"last updated must be a datetime, saw {last_updated}")
        self.last_updated = last_updated
        if errors:
            raise ValueError(f"Invalid PointDataUpdate: {', '.join(errors)}")

    def json(self):
        utc_ts_s = self.last_updated.replace(tzinfo=timezone.utc).timestamp()
        return {'id': self.point_id,
                'value': self.value,
                'last_updated': utc_ts_s * 1000}


class IngestStats(object):
    def __init__(self):
        self._points = []
        self._building = {}

    def summary(self, info):
        # infos, errors, num_points, sample_points, etc
        for k, v in info.items():
            if k == 'elapsed':
                self.elapsed(v)
            else:
                self._building[k] = v

    def add_points(self, points):
        self._points += points

    def elapsed(self, elapsed):
        self._building['processing_time_ms'] = math.floor(elapsed.total_seconds() * 1000)

    def json(self):
        return {
            'building': self._building,
            'points': self._points,
        }


@dataclass
class PointSelector:
    """A flexible interface to allow users to select sets of points"""
    # id, name or short_name
    orgs: List[Union[int, str]] = field(default_factory=list)
    # id or name
    buildings: List[Union[int, str]] = field(default_factory=list)

    # returned points are the superset of these three selectors
    point_ids: List[int] = field(default_factory=list)
    point_names: List[str] = field(default_factory=list)
    point_hashes: List[str] = field(default_factory=list)
    point_topics: List[str] = field(default_factory=list)

    # allow filtering out points w/o recent data
    updated_since: Optional[datetime] = None

    # PointType.id or PointType.tag_name
    point_types: List[Union[int, str]] = field(default_factory=list)

    # Equipment.id or Equipment.suffix
    equipment: List[Union[int, str]] = field(default_factory=list)
    # EquipmentType.id or EquipmentType.tag_name
    equipment_types: List[Union[int, str]] = field(default_factory=list)

    def json(self):
        ts = self.updated_since.timestamp() * 1000.0 if self.updated_since is not None else None
        dict = {k: getattr(self, k) for k in vars(self)
                if not k.startswith('__')}
        return {**dict, 'updated_since': ts}

    @staticmethod
    def from_json(dict):
        ps = PointSelector()
        for k in ps.__dict__.keys():
            val = dict.get(k, [])
            if k == 'updated_since':
                val = dict.get(k)
                if val is not None:
                    val = datetime.fromtimestamp(val / 1000.0)
            setattr(ps, k, val)
        return ps


@dataclass
class TimeseriesQuery:
    """Parameters needed to fetch timeseries data.

    Exactly one of point_ids or selector is required

    Note: the server may perform additional validation and reject queries
    which are constructable by the client

    For example values please refer to
        https://api.onboarddata.io/doc/#/buildings%3Aread/post_query_v2
    """
    start: datetime  # timezone required, year must be >= 2019
    end: datetime  # timezone required, year must be >= 2019
    selector: Optional[PointSelector] = None
    point_ids: List[int] = field(default_factory=list)

    units: List[str] = field(default_factory=list)

    @validator('point_ids')
    def points_or_selector_required(cls, point_ids, values):
        has_points = len(point_ids) > 0
        has_selector = values.get('selector') is not None
        if has_points == has_selector:
            raise ValueError("Exactly one of 'point_ids' or 'selector' is required")
        return point_ids

    @validator('start', 'end')
    def times_valid(cls, value, values):
        if value.year < 2019:
            raise ValueError(f'Time boundaries must be in 2019 or later, saw: {value}')
        if value.tzinfo is None:
            raise ValueError(f'Time boundaries require a timezone, saw: {value}')
        return value

    def json(self):
        return {
            'start': self.start.timestamp(),
            'end': self.end.timestamp(),
            'selector': self.selector.json() if self.selector is not None else None,
            'point_ids': self.point_ids,
            'units': self.units,
        }


@dataclass
class PointData:
    point_id: int
    raw: str
    unit: str
    columns: List[str]
    values: List[List[Union[str, float, int, None]]]



from setuptools import setup

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open("README.md", "r") as f:
    long_description = f.read()

setup(name='onboard.client',
      version='1.3.4',
      author='Nathan Merritt',
      author_email='nathan.merritt@onboarddata.io',
      description='Onboard API SDK',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/onboard-data/client-py',
      packages=['onboard.client'],
      install_requires=requirements,
      package_data={
        'onboard.client': ['py.typed'],
      },
      classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Topic :: Scientific/Engineering :: Information Analysis',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      python_requires='>=3.7',
      )



import urllib.parse
from datetime import datetime
import deprecation
from json import loads
from typing import List, Dict, Any, Optional, Tuple, Union, Iterator
from .util import divide_chunks, json
from .models import PointSelector, PointDataUpdate, IngestStats, \
    TimeseriesQuery, PointData
from .helpers import ClientBase
from .exceptions import OnboardApiException


class APIClient(ClientBase):
    def __init__(self, api_url, user=None, pw=None, api_key=None, token=None,
                 name='') -> None:
        super().__init__(api_url, user, pw, api_key, token, name)

    @json
    def whoami(self) -> Dict[str, str]:
        """returns the current account's information"""
        return self.get('/whoami')

    @json
    def get_account_actions(self) -> List[Dict[str, str]]:
        """returns the action audit log by or affecting the current account"""
        return self.get('/account-actions')

    @json
    def get_users(self) -> List[Dict[str, str]]:
        """returns the list of visible user accounts
          For organization admins this is all users in the organization
          For non-admin users this is just the current account
        """
        return self.get('/users')

    @json
    def get_organizations(self) -> List[Dict[str, str]]:
        return self.get('/organizations')

    @json
    def get_all_buildings(self) -> List[Dict[str, str]]:
        return self.get('/buildings')

    @json
    def get_tags(self) -> List[Dict[str, str]]:
        """returns a list of all the haystack tags in the system
        For more info, please see https://project-haystack.org/tag"""
        return self.get('/tags')

    @json
    def get_equipment_types(self) -> List[Dict[str, str]]:
        return self.get('/equiptype')

    @json
    def get_building_equipment(self, building_id: int) -> List[Dict[str, str]]:
        return self.get(f'/buildings/{building_id}/equipment?points=true')

    @json
    def get_equipment_by_ids(self, equipment_ids: List[int]) -> List[Dict[str, object]]:
        body = {'equipment_ids': equipment_ids}
        return self.post('/equipment/query', json=body)

    @json
    def get_building_changelog(self, building_id: int) -> List[Dict[str, object]]:
        """Returns a list of changelog entries for the specified building"""
        return self.get(f'/buildings/{building_id}/changelog')

    @json
    def select_points(self, selector: PointSelector) -> Dict[str, List[int]]:
        """returns point ids based on the provided selector"""
        return self.post('/points/select', json=selector.json())

    def check_data_availability(self, selector: PointSelector) -> Tuple[Optional[datetime], Optional[datetime]]:
        """Returns a tuple of data timestamps (most stale, most recent) for selected points"""
        @json
        def get_as_json():
            return self.post('/points/data-availability', json=selector.json())

        res = get_as_json()
        oldest = self.ts_to_dt(res['oldest'])
        newest = self.ts_to_dt(res['newest'])
        return (oldest, newest)

    def get_all_points(self) -> List[int]:
        """returns all points for all visible buildings"""
        buildings = self.get_all_buildings()
        point_ids = []
        for b in buildings:
            bldg_id = b['id']
            equipment = self.get_building_equipment(bldg_id)
            for e in equipment:
                point_ids += e['points']
        return point_ids

    def get_points_by_ids(self, point_ids: List[int]) -> List[Dict[str, str]]:
        @json
        def get_points(url):
            return self.get(url)

        points = []
        for chunk in divide_chunks(point_ids, 500):
            points_str = '[' + ','.join(str(id) for id in chunk) + ']'
            url = f'/points?point_ids={points_str}'
            try:
                points_chunk = get_points(url)
            except OnboardApiException as e:
                if '"status": 404' in str(e):
                    continue
                raise e
            points += points_chunk
        return points

    def get_points_by_datasource(self, datasource_hashes: List[str]) \
            -> List[Dict[str, str]]:
        datasource_hashes_chunked = list(divide_chunks(datasource_hashes, 125))

        @json
        def get_points(url):
            return self.get(url)

        points = []
        for chunk in datasource_hashes_chunked:
            hashes_str = "[" + ','.join([r"'" + c + r"'" for c in chunk]) + "]"
            query = urllib.parse.quote(hashes_str)
            url = f'/points?datasource_hashes={query}'
            try:
                points_chunk = get_points(url)
            except OnboardApiException as e:
                if '"status": 404' in str(e):
                    continue
                raise e
            points += points_chunk
        return points

    @json
    def get_all_point_types(self) -> List[Dict[str, str]]:
        return self.get('/pointtypes')

    @json
    def get_all_measurements(self) -> List[Dict[str, str]]:
        return self.get('/measurements')

    @json
    def get_all_units(self) -> List[Dict[str, str]]:
        return self.get('/unit')

    @deprecation.deprecated(deprecated_in="1.3.0",
                            details="Use stream_point_timeseries instead")
    @json
    def query_point_timeseries(self, point_ids: List[int],
                               start_time: Union[str, datetime],
                               end_time: Union[str, datetime]) -> List[Dict[str, Any]]:
        """Query a timespan for a set of point ids
        point_ids: a list of point ids
        start/end time: ISO formatted timestamp strings e.g. '2019-11-29T20:16:25Z' or a datetime
        """
        query = {
            'point_ids': point_ids,
            'start_time': self.dt_to_str(start_time),
            'end_time': self.dt_to_str(end_time),
        }
        return self.post('/query', json=query)

    def stream_point_timeseries(self, query: TimeseriesQuery) -> Iterator[PointData]:
        """Query a time interval for an explicit set of point ids or
        with a selector which describes which sensors to include.

        Example values docmentaed on the model tab here:
            https://api.onboarddata.io/doc/#/buildings%3Aread/post_query_v2
        """

        @json
        def query_call():
            return self.post('/query-v2', json=query.json(), stream=True,
                             headers={'Accept': 'application/x-ndjson'})
        query_call.raw_response = True

        with query_call() as res:
            for line in res.iter_lines():
                parsed = loads(line)
                yield PointData(**parsed)

    @json
    def update_point_data(self, updates: List[PointDataUpdate] = []) -> None:
        """Bulk update point data, returns the number of updated points
        updates: an iterable of models.PointDataUpdate objects"""
        for batch in divide_chunks(updates, 500):
            json = [u.json() for u in batch]
            self.post('/points_update', json=json)

    @json
    def send_ingest_stats(self, ingest_stats: IngestStats) -> None:
        """Send timing and diagnostic info to the portal
        ingest_stats: an instance of models.IngestStats"""
        json = ingest_stats.json()
        self.post('/ingest-stats', json=json)

    @json
    def get_ingest_stats(self) -> List[Dict[str, str]]:
        """returns ingest stats for all buildings"""
        return self.get('/ingest-stats')

    @json
    def get_alerts(self) -> List[Dict[str, str]]:
        """returns a list of active alerts for all buildings"""
        return self.get('/alerts')

    @json
    def copy_point_data(self, point_id_map: Dict[int, int],
                        start_time: Union[str, datetime],
                        end_time: Union[str, datetime]) -> str:
        """Copy data between points
        point_id_map: a map of source to destination point id
        start/end: ISO formatted timestamp strings e.g. '2019-11-29T20:16:25Z'
        returns: a string describing the operation
        """
        command = {
            'point_id_map': point_id_map,
            'start_time': self.dt_to_str(start_time),
            'end_time': self.dt_to_str(end_time),
        }
        return self.post('/point-data-copy', json=command)


class DevelopmentAPIClient(APIClient):
    def __init__(self, user=None, pw=None, api_key=None, token=None) -> None:
        super().__init__('https://devapi.onboarddata.io',
                         user, pw, api_key, token)


class ProductionAPIClient(APIClient):
    def __init__(self, user: Optional[str] = None, pw: Optional[str] = None,
                 api_key: Optional[str] = None,
                 token: Optional[str] = None) -> None:
        super().__init__('https://api.onboarddata.io',
                         user, pw, api_key, token)
